/*---------------------------------------------------------------------------
 * Interface driver for the GPCM 85230 ESCC DMA card. 
 *
 * Mods, 8254 & 85230-specific code by I1EPJ & IW1CIX
 * 1Mbps M. Vidmar RTX code (revised external clock mode) by IK4BLV 
 *---------------------------------------------------------------------------
 * GPCM ESCC card I/O ports allocations:
 *
 * Base+0  (e.g. 0x380) ESCC channel B control register
 * Base+1  (e.g. 0x381) ESCC channel B data register
 * Base+2  (e.g. 0x382) ESCC channel A control register
 * Base+3  (e.g. 0x383) ESCC channel A data register
 * Base+4  (e.g. 0x384) DMA & IRQ enable register:
 *                              bit 0: IRQ enable (1 enabled 0 disabled)
 *                              bit 1: DMA enable (1 enabled 0 disabled)
 * Base+8  (e.g. 0x388) 8254 Timer 0
 * Base+9  (e.g. 0x389) 8254 Timer 1
 * Base+10 (e.g. 0x38a) 8254 Timer 2
 * Base+11 (e.g. 0x38b) 8254 control register
 *
 * 8254 Out1 & Out2 go respectively to ESCC CTSA & CTSB pin.
 * When a timer expires an External Interrupt is generated.
 *
 * Channel A is DMA driven, channel B interrupt driven.
 *
 * See card schematic diagram for details.
 *---------------------------------------------------------------------------
 *
 * Heavily based on PI.C: 
 *
 * Interface driver for the VE3IFB 8530 card (PI card)
 * Copyright 1990 by Dave Perry, VE3IFB
 * Minor delinting - KA9Q 2/2/91 
 *
 * Portions of this driver were derived from the Eagle card driver by Art
 * Goldman, WA3CVG. It has been very extensively modified from his work, due
 * to the addition of DMA support and due to differences in the hardware. 
 * The PI card is NOT an Eagle card clone. It is an original design by Dave
 * Perry, VE3IFB.  Art's copyright notice follows: 
 *
 * Written by Art Goldman, WA3CVG - (c) Copyright 1987 All Rights Reserved
 * Permission for non-commercial use is hereby granted provided this notice
 * is retained.  For info call: (301) 997-3838. 
 *
 * and SCC.C: 
 *
 * Generic driver for Z8530 SCC chip in SLIP, KISS or AX.25 mode. 
 *
 * Written by R.E. Janssen (PE1CHL) using material from earlier EAGLE and
 * PC100 drivers in this package.
 *
 *
 */

#define FIFO                    /* use 85230 ESCC RX & TX FIFO enhancements */
#undef  FIFO_STATUS             /* use 85230 10x19 STATUS FIFO */
#undef  RECPROC                 /* use code to recover from ibuffails */

#include <time.h>
#include <stdio.h>
#include <dos.h>
#include <bios.h>
#include <ctype.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "pktdrvr.h"
#include "netuser.h"
#include "gpcmescc.h"
#include "85230.h"
#include "ax25.h"
#include "trace.h"
#include "pc.h"

#include "session.h"
#include "lapb.h"
#include "proc.h"
#include "ip.h"
#include "devparam.h"

#ifndef FP_OFF
#define FP_OFF(fp)      ((unsigned)(fp))
#endif
#ifndef FP_SEG
#define FP_SEG(fp)      ((unsigned)((unsigned long)(fp) >> 16))
#endif

/*
 * Message array 
 */
/* extern char **msg; */
extern int16 gpcm_acc_delay;
/*
 * Functions prototypes 
 */
static int32 gpcm_ctl __ARGS((struct iface * iface, int cmd, int set, int32 val));
static int gpcm_raw __ARGS((struct iface * iface, struct mbuf * bp));
static int gpcm_stop __ARGS((struct iface * iface));
static void rts __ARGS((struct gpcmchan * hp, int16 x));
static void setup_rx_dma __ARGS((struct gpcmchan * hp));
static void setup_tx_dma __ARGS((struct gpcmchan * hp, char *buffer, int length));
static void tdelay __ARGS((register struct gpcmchan * hp, unsigned int time));
static void a_rxint __ARGS((struct gpcmchan * hp));
static void a_txint __ARGS((struct gpcmchan * hp));
static void a_exint __ARGS((struct gpcmchan * hp));
static void b_rxint __ARGS((struct gpcmchan * hp));
static void b_txint __ARGS((struct gpcmchan * hp));
static void b_exint __ARGS((struct gpcmchan * hp));
static int scc_init __ARGS((register struct gpcmchan * hp));
extern void writescc __ARGS((int16 base, int16 ctl, int16 reg, int16 val));
extern char readscc __ARGS((int16 base, register int16 ctl, int16 reg));
extern void gpcm_set_acc_delay __ARGS((void));

/*
 * Device table - one entry per card 
 */
static struct GPCMTAB GPCM[GPCMMAX];

/*
 * handler interrupt vector table 
 */
static INTERRUPT(*gpcmhandle[]) () =
{
	 gpcm0vec,
	 gpcm1vec,
	 gpcm2vec
};

static int16 Page_regs[] = {
			    0x87, 0x83, 0x81, 0x82, 0, 0x8b, 0x89, 0x8a
};

/*
 * channel table - 2 entries per card 
 */
static struct gpcmchan GPCMchan[2 * GPCMMAX];

/*
 * Global variables 
 */
static int16 gpcmnbr, have_escc, have_scc;
static int16 auto_enab = AUTO_ENAB;
static int16 rxfifolevel = RFIL;
static int16 txfifolevel = TFIL;

/*
 * Setup 8253/8254 chip for time delay 
 */
static void tdelay(hp, time)
register struct gpcmchan *hp;
unsigned int time;              /* Time to delay in milliseconds */
{
	int n, port;
	unsigned int t1;
	unsigned char sc;

	if (hp->base & 2) {     /* If A channel */
		sc = SC1;
		port = hp->cardbase + TMR1;
	} else {
		sc = SC2;
		port = hp->cardbase + TMR2;
		writescc(hp->cardbase, hp->base + CTL, R1, EXT_INT_ENAB);
	}

	t1 = 4 * time;          /* make 250 us units */

	/*
	 * Setup timer sc 
	 */
	outportb(hp->cardbase + TMRCMD, sc | LSB_MSB | MODE0);

	/*
	 * satisfy 8253 access time restriction 
	 */
	for (n = 0; n < 5; n++);

	/*
	 * Setup time delay, low byte 
	 */
	outportb(port, t1 & 0xFF);

	/*
	 * satisfy 8253 access time restriction 
	 */
	for (n = 0; n < 5; n++);

	/*
	 * Setup time delay, high byte 
	 */
	outportb(port, (t1 >> 8) & 0xFF);

	/*
	 * Enable correct int for timeout 
	 */
	writescc(hp->cardbase, hp->base + CTL, R15, CTSIE);
	writescc(hp->cardbase, hp->base + CTL, R0, RES_EXT_INT);
}

/*
 * Master interrupt handler.  One interrupt at a time is handled here.
 * Service routines are called from here. 
 */
void gpcmint(dev)
int dev;
{
	register char st;
	register int16 pcbase;
	struct gpcmchan *hp;

	GPCM[dev].ints++;
	pcbase = GPCM[dev].addr;

	/*
	 * Read interrupt status register (only valid from channel A).
	 * Process all pending interrupts in while loop.
	 *
	 * The if order MUST MATCH with the interrupts priority list;
	 * otherwise, the RES_H_IUS command issued at the end may
	 * reset THE WRONG INTERRUPT and the while will loop forever !!
	 *
	 * From Zilog 85230 data sheet the 85230 interrupt priority is
	 * (from highest to lowest): 
	 * Channel A RX, Channel A TX, Channel A EXT,
	 * Channel B RX, Channel B TX, Channel B EXT.
	 */
	hp = &GPCMchan[2 * dev];/* Assume channel A */
	while ((st = readscc(hp->cardbase, pcbase + CHANA + CTL, R3)) != 0) {

		if (st & CHARxIP) {

			/*
			 * Channel A Rcv Interrupt Pending 
			 */
			hp = &GPCMchan[2 * dev];
			a_rxint(hp);
		} else if (st & CHATxIP) {

			/*
			 * Channel A Transmit Int Pending 
			 */
			hp = &GPCMchan[2 * dev];
			a_txint(hp);
		} else if (st & CHAEXT) {

			/*
			 * Channel A External Status Int 
			 */
			hp = &GPCMchan[2 * dev];
			a_exint(hp);
		} else if (st & CHBRxIP) {

			/*
			 * Channel B Rcv Interrupt Pending 
			 */
			hp = &GPCMchan[(2 * dev) + 1];
			b_rxint(hp);
		} else if (st & CHBEXT) {

			/*
			 * Channel B External Status Int 
			 */
			hp = &GPCMchan[(2 * dev) + 1];
			b_exint(hp);
		} else if (st & CHBTxIP) {

			/*
			 * Channel B Transmit Int Pending 
			 */
			hp = &GPCMchan[(2 * dev) + 1];
			b_txint(hp);
		}

		/*
		 * Reset highest interrupt under service 
		 */
		writescc(hp->cardbase, hp->base + CTL, R0, RES_H_IUS);
	}
}

static void a_exint(hp)
register struct gpcmchan *hp;
{
	register int16 cmd;
	char st, i_state;
	int length;
	int32 t, ca;

	/*
	 * disable interrupts 
	 */
	i_state = dirps();

	/*
	 * Fetch status 
	 */
	st = readscc(hp->cardbase, hp->base + CTL, R0);

	/*
	 * reset external status latch 
	 */
	writescc(hp->cardbase, CTL + hp->base, R0, RES_EXT_INT);
	cmd = hp->base + CTL;
	hp->exints++;

	if ((hp->rstate >= ACTIVE) && (st & BRK_ABRT)) {
		setup_rx_dma(hp);
		hp->rstate = ACTIVE;
	}
	switch (hp->tstate) {
		case ACTIVE:
			free_p(hp->sndbuf);
			hp->sndbuf = NULLBUF;
			hp->tstate = FLAGOUT;
			tdelay(hp, hp->squeldelay);
			break;
		case FLAGOUT:
			if ((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF) {

				/*
				 * Nothing to send - return to receive mode 
				 */
				hp->tstate = IDLE;
				rts(hp, OFF);
				restore(i_state);
				return;
			}
			/* NOTE - fall through if more to send */
		case ST_TXDELAY:

			/*
			 * Disable DMA chan 
			 */
			outportb(DMA_MASK, DMA_DISABLE | hp->dmachan);

			/*
			 * Set up for TX dma 
			 */
			writescc(hp->cardbase, cmd, R1, WT_FN_RDYFN | EXT_INT_ENAB);

			/*
			 * Get all chars 
			 */
			length = pullup(&hp->sndbuf, hp->txdmabuf, hp->bufsiz);

			/*
			 * Setup DMA controller for tx 
			 */
			setup_tx_dma(hp, hp->txdmabuf, length);

			/*
			 * Select transmit interrupts to enable, allow DMA on
			 * chan 
			 */
			outportb(DMA_MASK, DMA_ENABLE | hp->dmachan);

			/*
			 * reset CRC, Txint pend 
			 */
			writescc(hp->cardbase, cmd, R0, RES_Tx_CRC | RES_Tx_P);

			/*
			 * allow Underrun int only 
			 */
			writescc(hp->cardbase, cmd, R15, TxUIE);

			/*
			 * Enable TX DMA 
			 */
			writescc(hp->cardbase, cmd, R1, WT_RDY_ENAB | WT_FN_RDYFN | EXT_INT_ENAB);

			/*
			 * Send CRC on underrun 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EOM_L);

			/*
			 * packet going out now 
			 */
			hp->tstate = ACTIVE;
			break;
#if 0
		case WAITTIME:
#endif
		case DEFER:

			/*
			 * we have deferred prev xmit attempt See Intel
			 * Microcommunications Handbook, p2-308 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			if ((readscc(hp->cardbase, cmd, R0) & DCD) != 0) {
				hp->tstate = DEFER;
				tdelay(hp, 100);

				/*
				 * Defer until dcd transition or 100mS
				 * timeout 
				 */
				writescc(hp->cardbase, CTL + hp->base, R15, CTSIE | DCDIE);
				restore(i_state);
				return;
			}

			/*
			 * DCD not active, roll p-persist dice 
			 */
			if ((rand() & 0xff) > uchar(hp->persist)) {

				/*
				 * We have lost, so defer TX attempt for
				 * slottime 
				 */
				hp->tstate = DEFER;
				tdelay(hp, hp->slotime);
				restore(i_state);
				return;
			}

			/*
			 * We have won, so turn on transmitter and set timer
			 * for txdelay 
			 */
			rts(hp, ON);
			hp->tstate = ST_TXDELAY;
			tdelay(hp, hp->txdelay);
			restore(i_state);
			return;
	}                       /* switch(hp->tstate) */

	restore(i_state);
}                               /* a_exint() */

/*
 * Receive interrupt handler for the A channel 
 */
static void a_rxint(hp)
register struct gpcmchan *hp;
{
	register int16 cmd;
	register int16 bytecount;
	char rse, i_state;
	struct mbuf *bp;
	struct phdr phdr;

	i_state = dirps();      /* disable interrupts */
	hp->rxints++;
	cmd = hp->base + CTL;

	/*
	 * Get special condition bits from R1 
	 */
	rse = readscc(hp->cardbase, cmd, R1);

	if (rse & Rx_OVR) {

		/*
		 * Receiver overrun 
		 */
		hp->rovers++;
		hp->rstate = RXERROR;
	}
	if (rse & END_FR) {

		/*
		 * End of frame figure length of frame from 8237 
		 */
		outportb(DMA_RESETFF, 0);       /* reset firstlast ff */
		bytecount = inportb(hp->dma_wcr);
		bytecount += inportb(hp->dma_wcr) << 8;

		/*
		 * Allow for the extra space for phdr 
		 */
		bytecount = (hp->bufsiz - 1 - sizeof(struct phdr)) - bytecount;

		/*
		 * Check for frames too short and CRC error 
		 */
		if ((rse & CRC_ERR) || (hp->rstate > ACTIVE) || (bytecount < 10)) {
			if ((bytecount >= 10) && (rse & CRC_ERR))
				hp->crcerr++;

			/*
			 * Reset buffer pointers 
			 */
			hp->rstate = ACTIVE;
			setup_rx_dma(hp);
		} else {

			/*
			 * Here we have a valid frame Toss 2 crc bytes 
			 */
			hp->rcvbuf->cnt = bytecount - 2;
			bp = pushdown(hp->rcvbuf, sizeof(phdr));
			phdr.type = CL_AX25;
			phdr.iface = hp->iface;
			memcpy(bp->data, (char *) &phdr, sizeof(phdr));
			hp->rcvbuf = NULLBUF;
			enqueue(&Hopper, bp);   /* queue it in */
			hp->rxframes++;

			/*
			 * packet queued - get buffer for next frame 
			 */
			hp->rcvbuf = alloc_mbuf(hp->bufsiz + sizeof(phdr));
			if (hp->rcvbuf != NULLBUF)
				/* Allow room for phdr */
				hp->rcvbuf->data += sizeof(struct phdr);
			setup_rx_dma(hp);
		}               /* end good frame queued */
	}                       /* end EOF check */
	writescc(hp->cardbase, hp->base + CTL, R0, ERR_RES);    /* error reset */
	restore(i_state);
}

static void a_txint(hp)
register struct gpcmchan *hp;
{
	register int16 cmd;
	char i_state;
	int32 t, ca;

	i_state = dirps();
	cmd = CTL + hp->base;
	hp->txints++;

	switch (hp->tstate) {
		case IDLE:
			/* Transmitter idle. Find a frame for transmission */
			if ((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF) {
				rts(hp, OFF);
				restore(i_state);
				return;
			}

			/*
			 * If a buffer to send, we drop thru here 
			 */
#if 0
		case WAITTIME:
			hp->tstate = DEFER;
			tdelay(hp, hp->waittime);
			restore(i_state);
			return;
#endif
		case DEFER:

			/*
			 * we may have deferred prev xmit attempt 
			 *
			 * Check DCD - debounce it See Intel Microcommunications
			 * Handbook, p2-308 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			if ((readscc(hp->cardbase, cmd, R0) & DCD) != 0) {
				hp->tstate = DEFER;
				tdelay(hp, 100);

				/*
				 * DCD active, defer until DCD transition or
				 * timeout 
				 */
				writescc(hp->cardbase, cmd, R15, CTSIE | DCDIE);
				restore(i_state);
				return;
			}

			/*
			 * DCD not active, roll p-persist dice 
			 */
			if ((rand() & 0xff) > uchar(hp->persist)) {

				/*
				 * We have lost, so defer TX attempt for
				 * slottime 
				 */
				hp->tstate = DEFER;
				tdelay(hp, hp->slotime);
				restore(i_state);
				return;
			}

			/*
			 * We have won, so turn on transmitter and set timer
			 * for txdelay 
			 */
			rts(hp, ON);
			hp->tstate = ST_TXDELAY;
			tdelay(hp, hp->txdelay);
			restore(i_state);
			return;
		default:
			break;
	}                       /* end switch(hp->state) */

	restore(i_state);
}                               /* a_txint */

static void b_rxint(hp)
register struct gpcmchan *hp;
{
	register int16 cmd;
	char rse, i_state;
	struct mbuf *bp;
	struct phdr phdr;

	i_state = dirps();      /* disable interrupts */
	hp->rxints++;
	cmd = CTL + hp->base;

	while ((readscc(hp->cardbase, cmd, R0)) & Rx_CH_AV) {

		/*
		 * Interrupt hits when there are at least 4 chars in the FIFO
		 * Read special condition bits before reading the data char 
		 */
		rse = readscc(hp->cardbase, cmd, R1);   /* get status byte from
							 * R1 */
		if (rse & Rx_OVR) {

			/*
			 * Rx overrun - toss buffer, reset buffer pointers 
			 */
			hp->rcp = hp->rcvbuf->data;
			hp->rcvbuf->cnt = 0;
			hp->rstate = RXERROR;   /* set error flag */
			hp->rovers++;
		} else if (hp->rcvbuf->cnt >= hp->bufsiz - sizeof(struct phdr)) {

			/*
			 * Too large -- toss buffer, reset buffer pointers 
			 */
			hp->rcp = hp->rcvbuf->data;
			hp->rcvbuf->cnt = 0;
			hp->rstate = TOOBIG;    /* when set, chars are not
						 * stored */
		}
		if (hp->rstate == ACTIVE) {

			/*
			 * If no errors, read character from ESCC 
			 */
			*hp->rcp++ = readscc(hp->cardbase, cmd, R8);
			hp->rcvbuf->cnt++;
		} else {
			/* got to empty FIFO */
			/* (void) readscc(hp->cardbase, cmd, R8); */

			/*
			 * reset err latch 
			 */
			writescc(hp->cardbase, cmd, R0, ERR_RES);

			hp->rstate = ACTIVE;
		}
	}
	if (rse & END_FR) {

		/*
		 * END OF FRAME -- Make sure Rx was active 
		 */
		if (hp->rcvbuf->cnt > 0) {
			if ((rse & CRC_ERR) || (hp->rstate > ACTIVE) || (hp->rcvbuf->cnt < 10)) {
				if ((hp->rcvbuf->cnt >= 10) && (rse & CRC_ERR))
					hp->crcerr++;   /* Ignore noise */

				hp->rcp = hp->rcvbuf->data;
				hp->rcvbuf->cnt = 0;
			} else {

				/*
				 * Here we have a valid frame 
				 */
				hp->rcvbuf->cnt -= 2;   /* Toss 2 crc bytes */
				bp = pushdown(hp->rcvbuf, sizeof(phdr));
				phdr.type = CL_AX25;
				phdr.iface = hp->iface;
				memcpy(bp->data, (char *) &phdr, sizeof(phdr));
				enqueue(&Hopper, bp);   /* queue it in */
				hp->rxframes++;

				/*
				 * packet queued - get buffer for next frame 
				 */
				hp->rcvbuf = alloc_mbuf(hp->bufsiz + sizeof(struct phdr));
				if (hp->rcvbuf == NULLBUF) {

					/*
					 * No memory - abort rx 
					 */
					writescc(hp->cardbase, cmd, R3, Rx8);
					restore(i_state);
					return;
				}
				hp->rcvbuf->data += sizeof(struct phdr);
				hp->rcp = hp->rcvbuf->data;
				hp->rcvbuf->cnt = 0;
			}       /* end good frame queued */
		}               /* end check for active receive upon EOF */
		hp->rstate = ACTIVE;    /* and clear error status */
	}                       /* end EOF check */
	restore(i_state);
}

static void b_txint(hp)
register struct gpcmchan *hp;
{
	register int16 cmd, n;
	char i_state;
	int c;

	i_state = dirps();
	cmd = CTL + hp->base;

	if (hp->tstate != DEFER && hp->tstate)
		hp->txints++;

	switch (hp->tstate) {
		case CRCOUT:
			hp->tstate = FLAGOUT;
			tdelay(hp, hp->squeldelay);
			restore(i_state);
			return;
		case IDLE:

			/*
			 * Transmitter idle. Find a frame for transmission 
			 */
			if ((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF) {

				/*
				 * Nothing to send - return to receive mode
				 * Tx OFF now - flag should have gone 
				 */
				rts(hp, OFF);
				restore(i_state);
				return;
			}

			/*
			 * If a buffer to send, we drop thru here 
			 */
#if 0
		case WAITTIME:
			hp->tstate = DEFER;
			tdelay(hp, hp->waittime);
			restore(i_state);
			return;
#endif
		case DEFER:

			/*
			 * We may have deferred prev xmit attempt 
			 *
			 * Check DCD - debounce it See Intel Microcommunications
			 * Handbook, p2-308 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			if ((readscc(hp->cardbase, cmd, R0) & DCD) != 0) {
				hp->tstate = DEFER;
				tdelay(hp, 100);

				/*
				 * defer until DCD transition or 100 ms
				 * timeout 
				 */
				writescc(hp->cardbase, cmd, R15, CTSIE | DCDIE);
				restore(i_state);
				return;
			}

			/*
			 * DCD not active, roll p-persist dice 
			 */
			if ((rand() & 0xff) > uchar(hp->persist)) {

				/*
				 * We have lost, so defer TX attempt for
				 * slottime 
				 */
				hp->tstate = DEFER;
				tdelay(hp, hp->slotime);
				restore(i_state);
				return;
			}

			/*
			 * We have won, so turn on transmitter and set timer
			 * for txdelay 
			 */
			rts(hp, ON);
			hp->tstate = ST_TXDELAY;
			tdelay(hp, hp->txdelay);
			restore(i_state);
			return;

		case ACTIVE:

			/*
			 * Here we are actively sending a frame 
			 *
			 * Write enough characters to fill out transmit FIFO
			 * Next interrupt will occur:
			 *  if txfifolevel != 0: when FIFO is empty again
			 *  if txfifolevel == 0: when FIFO has top byte free
			 */
			for (n = 0; n < 4; n++) {
				c = PULLCHAR(&hp->sndbuf);
				if (c == -1)
					break;
				writescc(hp->cardbase, cmd, R8, c);
				/*
				 * Exit when TX FIFO full
				 */
				if ((readscc(hp->cardbase, cmd, R0) & Tx_BUF_EMP) == 0)
					break;

				/*
				 * stuffing a char satisfies Interrupt
				 * condition 
				 */
			}
			if (c == -1) {

				/*
				 * No more characters to send 
				 */
				free_p(hp->sndbuf);
#ifdef notdef
/* is that really needed ?
 * A underrun does not automagically triggers an external interrupt ??
 */
				if ((readscc(hp->cardbase, cmd, R0) & 0x40)) {

					/*
					 * Unexpected underrun ? 
					 */
					hp->tunders++;
					writescc(hp->cardbase, cmd, R0, SEND_ABORT);
					hp->tstate = FLAGOUT;
					tdelay(hp, hp->squeldelay);
					restore(i_state);
					return;
				}
#endif
				/*
				 * Now we expect to underrun 
				 */
				hp->tstate = UNDERRUN;

				/*
				 * Send flags on underrun 
				 */
				if (hp->speed) {
					writescc(hp->cardbase, cmd, R10, CRCPS | NRZI);
				} else {
					writescc(hp->cardbase, cmd, R10, CRCPS | NRZI);
				}

				/*
				 * reset Tx Int Pend 
				 */
				writescc(hp->cardbase, cmd, R0, RES_Tx_P);
			}
			restore(i_state);
			return;
	}
	restore(i_state);
}

/*
 * GPCM ESCC External/Status interrupts (for the B channel) This can be
 * caused by a receiver abort, or a Tx UNDERRUN/EOM. Receiver automatically
 * goes to Hunt on an abort. 
 *
 * If the Tx Underrun interrupt hits, change state and issue a reset command for
 * it, and return. 
 */
static void b_exint(hp)
register struct gpcmchan *hp;
{
	char st, i_state, n;
	register int16 cmd;
	char c;

	cmd = CTL + hp->base;
	i_state = dirps();      /* disable interrupts */
	hp->exints++;

	/*
	 * Fetch status 
	 */
	st = readscc(hp->cardbase, cmd, R0);

	/*
	 * reset external status latch 
	 */
	writescc(hp->cardbase, cmd, R0, RES_EXT_INT);


	switch (hp->tstate) {
		case ACTIVE:

			/*
			 * Unexpected underrun 
			 */
			free_p(hp->sndbuf);
			writescc(hp->cardbase, cmd, R0, SEND_ABORT);
			hp->tstate = FLAGOUT;
			hp->tunders++;
			tdelay(hp, hp->squeldelay);
			restore(i_state);
			return;
		case UNDERRUN:
			hp->tstate = CRCOUT;
			restore(i_state);
			return;
		case FLAGOUT:

			/*
			 * Find a frame for transmission 
			 */
			if ((hp->sndbuf = dequeue(&hp->sndq)) == NULLBUF) {

				/*
				 * Nothing to send - return to receive mode
				 * Tx OFF now - flag should have gone 
				 */
				rts(hp, OFF);
				hp->tstate = IDLE;
				restore(i_state);
				return;
			}

			/*
			 * Reset CRC for next frame 
			 */
			writescc(hp->cardbase, cmd, R0, RES_Tx_CRC);

			/*
			 * Send abort on underrun 
			 */
			if (hp->speed) {
				writescc(hp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			} else {
				writescc(hp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			}

			/*
			 * Fill out transmit FIFO 
			 */
			for (n = 0; n < 4; n++) {
				c = PULLCHAR(&hp->sndbuf);
				writescc(hp->cardbase, cmd, R8, c);
			}

			/*
			 * Reset end of message latch 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EOM_L);

			/*
			 * Select transmit interrupts to enable 
			 */
			writescc(hp->cardbase, cmd, R15, TxUIE);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R1, TxINT_ENAB | EXT_INT_ENAB);

			hp->tstate = ACTIVE;    /* char going out now */
			restore(i_state);
			return;
#if 0
		case WAITTIME:
#endif

		case DEFER:

			/*
			 * Check DCD - debounce it See Intel
			 * Microcommunications Handbook, p2-308 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			if ((readscc(hp->cardbase, cmd, R0) & DCD) != 0) {
				hp->tstate = DEFER;
				tdelay(hp, 100);

				/*
				 * defer until DCD transition or timeout 
				 */
				writescc(hp->cardbase, cmd, R15, CTSIE | DCDIE);
				restore(i_state);
				return;
			}

			/*
			 * DCD not active, roll p-persist dice 
			 */
			if ((rand() & 0xff) > uchar(hp->persist)) {

				/*
				 * We have lost, so defer TX attempt for
				 * slottime 
				 */
				hp->tstate = DEFER;
				tdelay(hp, hp->slotime);
				restore(i_state);
				return;
			}

			/*
			 * We have won, so turn on transmitter and set timer
			 * for txdelay 
			 */
			rts(hp, ON);
			hp->tstate = ST_TXDELAY;
			tdelay(hp, hp->txdelay);
			restore(i_state);
			return;

		case ST_TXDELAY:

			writescc(hp->cardbase, cmd, R0, RES_Tx_CRC);    /* reset for next frame */

			/*
			 * Send abort on underrun 
			 */
			if (hp->speed) {
				writescc(hp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			} else {
				writescc(hp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			}

			/*
			 * Fill out transmit FIFO 
			 */
			for (n = 0; n < 4; n++) {
				c = PULLCHAR(&hp->sndbuf);
				writescc(hp->cardbase, cmd, R8, c);
			}

			/*
			 * Reset end of message latch 
			 */
			writescc(hp->cardbase, cmd, R0, RES_EOM_L);

			/*
			 * select transmit interrupts to enable 
			 */
			writescc(hp->cardbase, cmd, R15, TxUIE);
			writescc(hp->cardbase, cmd, R0, RES_EXT_INT);
			writescc(hp->cardbase, cmd, R1, TxINT_ENAB | EXT_INT_ENAB);

			hp->tstate = ACTIVE;    /* char going out now */
			restore(i_state);
			return;
	}

	/*
	 * Receive Mode only This triggers when hunt mode is entered, & since
	 * an ABORT automatically enters hunt mode, we use that to clean up
	 * any waiting garbage 
	 */
	if ((hp->rstate == ACTIVE) && (st & BRK_ABRT)) {
		for (n = 0; n < 8; n++) {
			(void) readscc(hp->cardbase, cmd, R8);
		}
		hp->rcp = hp->rcvbuf->data;
		hp->rcvbuf->cnt = 0;    /* rewind on DCD transition */
	}
	restore(i_state);
}

/*
 * SET Transmit or Receive Mode Set RTS (request-to-send) to modem on
 * Transmit 
 */
static void rts(hp, x)
register struct gpcmchan *hp;
int16 x;
{
	int16 tc;
	long br;
	int16 cmd;

	cmd = CTL + hp->base;

	if (x == ON) {

		/*
		 * Turn Tx ON and Receive OFF Exints off first to avoid abort
		 * int 
		 */
		writescc(hp->cardbase, cmd, R15, 0);
		writescc(hp->cardbase, cmd, R3, Rx8);   /* Rx off */
		hp->rstate = IDLE;
		if (cmd & 2) {  /* if channel a */

			/*
			 * Set up for TX dma 
			 */
			writescc(hp->cardbase, cmd, R1, WT_FN_RDYFN | EXT_INT_ENAB);
		} else {

			/*
			 * No interrupts 
			 */
			writescc(hp->cardbase, cmd, R1, 0);
		}
		writescc(hp->cardbase, cmd, R5, TxCRC_ENAB | RTS | TxENAB | Tx8 | DTR);

		/*
		 * Transmitter now on 
		 */
	} else {

		/*
		 * Tx OFF and Rx ON 
		 */
		hp->tstate = IDLE;
		writescc(hp->cardbase, cmd, R5, Tx8 | DTR);     /* TX off */

		/*
		 * Turn on the receiver and hunt for a flag Use DCD-enabled
		 * RX to avoid tons of interrupts 
		 */
		writescc(hp->cardbase, cmd, R3, RxENABLE | RxCRC_ENAB | Rx8 | auto_enab);
		hp->rstate = ACTIVE;    /* Normal state */

		if (cmd & 2) {  /* if channel a */
			setup_rx_dma(hp);
		} else {
			/* reset error bits */
			/* writescc(hp->cardbase,cmd,R0,ERR_RES); */
			/* reset buffer pointers */
			hp->rcp = hp->rcvbuf->data;
			hp->rcvbuf->cnt = 0;
			writescc(hp->cardbase, cmd, R1, (INT_ALL_Rx | EXT_INT_ENAB));
		}
#ifdef FIFO_STATUS
		writescc(hp->cardbase, cmd, R15, BRKIE | FIFOEN);       /* allow ABORT int */
#else
		writescc(hp->cardbase, cmd, R15, BRKIE);        /* allow ABORT int */
#endif
	}
}

static void setup_rx_dma(hp)
register struct gpcmchan *hp;
{
	unsigned buf_offset, buf_segment;
	int cmd;
	long longseg, dma_abs, dma_page;
	char i_state;
	
	/* 
	 * disable interrupts
	 */
	i_state = dirps();

	cmd = hp->base + CTL;

	if (!hp->rcvbuf) {

		/*
		 * No rx buffer available 
		 */
		restore(i_state);
		return;
	}

	/*
	 * Calculate high order 4 bits of the buffer area and store them in
	 * the DMA page register 
	 */
	buf_offset = FP_OFF(hp->rcvbuf->data);
	buf_segment = FP_SEG(hp->rcvbuf->data);
	longseg = (long) buf_segment;
	dma_abs = (longseg << 4) + (long) buf_offset;
	dma_page = dma_abs >> 16;

	if (((dma_abs + hp->bufsiz - 1) >> 16) != dma_page)
		tprintf("ERROR-RX DMA page boundatory violation\n");

	/*
	 * Get ready for RX DMA 
	 */
	writescc(hp->cardbase, cmd, R1, WT_FN_RDYFN | WT_RDY_RT | INT_ERR_Rx | EXT_INT_ENAB);
	outportb(DMA_MASK, DMA_DISABLE | hp->dmachan);  /* Disable DMA chan */

	/*
	 * Set DMA mode register to single transfers, incrementing address,
	 * auto init, writes 
	 */
	outportb(DMA_MODE, DMA_RX_MODE | hp->dmachan);

	outportb(hp->page_addr, dma_page);      /* Store in  64K DMA page */
	outportb(DMA_RESETFF, 0);       /* reset byte pointer flipflop */

	/*
	 * Output buffer start (dest) address 
	 */
	outportb(hp->dma_dest, dma_abs);
	outportb(hp->dma_dest, dma_abs >> 8);

	/*
	 * output DMA maximum byte count 
	 */
	outportb(hp->dma_wcr, hp->bufsiz - 1 - sizeof(struct phdr));
	outportb(hp->dma_wcr, (hp->bufsiz - 1 - sizeof(struct phdr)) >> 8);

	/*
	 * Unmask channel 1 (start DMA) 
	 */
	outportb(DMA_MASK, DMA_ENABLE | hp->dmachan);   /* Enable DMA chan */

	/*
	 * If a packet is already coming in, this line is supposed to mess up
	 * the crc to avoid receiving a partial packet 
	 */
	writescc(hp->cardbase, cmd, R0, RES_Rx_CRC);

	/*
	 * Enable RX dma 
	 */
	writescc(hp->cardbase, cmd, R1, WT_RDY_ENAB | WT_FN_RDYFN | WT_RDY_RT | INT_ERR_Rx | EXT_INT_ENAB);

	restore(i_state);
}

static void setup_tx_dma(hp, buffer, length)
struct gpcmchan *hp;
char *buffer;
int length;
{
	unsigned buf_offset, buf_segment;
	long longseg, dma_abs, dma_page;
	char i_state;

	i_state = dirps();      /* disable interrupts */

	/*
	 * Calculate high order 4 bits of the buffer area and store them in
	 * the DMA page register 
	 */

	buf_offset = FP_OFF(buffer);
	buf_segment = FP_SEG(buffer);
	longseg = (long) buf_segment;
	dma_abs = (longseg << 4) + (long) buf_offset;
	dma_page = dma_abs >> 16;

	outportb(DMA_MASK, DMA_DISABLE | hp->dmachan);  /* Disable DMA chan */
	if (((dma_abs + length) >> 16) != dma_page)
		tprintf("ERROR-TX DMA page boundatory violation\n");
	--length;               /* Adjust length for DMA chip */

	/*
	 * Set DMA mode register to single transfers, incrementing address,
	 * no auto init, reads 
	 */
	outportb(DMA_MODE, DMA_TX_MODE | hp->dmachan);

	outportb(hp->page_addr, dma_page);      /* Store in 64K DMA page */
	outportb(DMA_RESETFF, 0);       /* reset byte pointer flipflop */
	outportb(hp->dma_dest, dma_abs);        /* Output buffer start
						 * (source) address */
	outportb(hp->dma_dest, dma_abs >> 8);

	/*
	 * output byte count 
	 */
	outportb(hp->dma_wcr, length);
	outportb(hp->dma_wcr, (length) >> 8);

	restore(i_state);
}

/* Initialize GPCM ESCC CARD controller parameters */
static int scc_init(register struct gpcmchan * hp)
{
	int16 tc;
	long br;
	char i_state;
	register int16 cmd;

	/*
	 * Initialize 85230 channel for SDLC operation 
	 */
	cmd = CTL + hp->base;
	i_state = dirps();

	switch (cmd & 2) {
		case 2:

			/*
			 * Reset channel A 
			 */
			writescc(hp->cardbase, cmd, R9, CHRA);

			/*
			 * Initialize interrupt vector 
			 */
			writescc(hp->cardbase, cmd, R2, 0xff);
			break;
		case 0:

			/*
			 * Reset channel B 
			 */
			writescc(hp->cardbase, cmd, R9, CHRB);
			break;
	}

	/*
	 * Deselect all Rx and Tx interrupts 
	 */
	writescc(hp->cardbase, cmd, R1, 0);

	/*
	 * Turn off external interrupts (like CTS/CD) 
	 */
	writescc(hp->cardbase, cmd, R15, 0);

	/*
	 * X1 clock, SDLC mode 
	 */
	writescc(hp->cardbase, cmd, R4, SDLC | X1CLK);

	/*
	 * Now some misc Tx/Rx parameters 
	 *
	 * CRC PRESET 1, NRZI Mode 
	 */
	if (hp->speed) {
		writescc(hp->cardbase, cmd, R10, CRCPS | NRZI);

		/*
		 * Tx Clk from DPLL (=internal ESCC /32 divisor) Rcv Clk from
		 * DPLL (the true DPLL) TRxC pin outputs DPLL 
		 */
		writescc(hp->cardbase, cmd, R11, TCDPLL | RCDPLL | TRxCDP | TRxCOI);
	} else {
		writescc(hp->cardbase, cmd, R10, CRCPS | NRZI);

		/*
		 * Tx Clk from Trxcl. Rcv Clk from Rtxcl, TRxC pin is input 
		 */
	      /* writescc(hp->cardbase, cmd, R11, TCTRxCP); */
	      /* Modifica mia */
		 writescc(hp->cardbase, cmd, R11, RTxCX | RCRTxCP | TCRTxCP);
	}

	/*
	 * Null out SDLC start address 
	 */
	writescc(hp->cardbase, cmd, R6, 0);

	/*
	 * SDLC flag 
	 */
	writescc(hp->cardbase, cmd, R7, FLAG);

	/*
	 * Set up the Transmitter but don't enable it DTR, 8 bit TX chars
	 * only - TX NOT ENABLED 
	 */
	writescc(hp->cardbase, cmd, R5, Tx8 | DTR);

	/*
	 * Receiver - intial setup only - more later 
	 */
	writescc(hp->cardbase, cmd, R3, Rx8);   /* 8 bits/char */

	/*
	 * Setting up BRG now - turn it off first 
	 */
	writescc(hp->cardbase, cmd, R14, 0);    /* BRG off */

	/*
	 * set the 32x time constant for the BRG in Receive mode 
	 */
	if (hp->speed) {
		br = hp->speed; /* get desired speed */
		tc = (XTAL / (32 * (2 * br))) - 2;      /* calc 32X BRG divisor */
	} else {
		tc = 0;
	}

	writescc(hp->cardbase, cmd, R12, tc & 0xFF);    /* lower byte */
	writescc(hp->cardbase, cmd, R13, (tc >> 8) & 0xFF);     /* upper byte */

	/*
	 * Following subroutine sets up and ENABLES the receiver 
	 */
	rts(hp, OFF);           /* TX OFF and RX ON */

	if (hp->speed) {

		/*
		 * DPLL frm BRG, BRG src RTxC 
		 */
		writescc(hp->cardbase, cmd, R14, SSBR);
		
	} else {

		/*
		 * DPLL frm rtxc,BRG src RTxC 
		 */
		writescc(hp->cardbase, cmd, R14, DISDPLL | SNRZI);
	}
	/* Queste righe le ho messe sopra che a 38400 funzionano per 1m2 da testare */
	writescc(hp->cardbase, cmd, R14, SEARCH);       /* SEARCH mode, keep BRG
							 * source */
	writescc(hp->cardbase, cmd, R14, BRENABL);      /* Enable the BRG */

	if (!(cmd & 2)) {       /* if channel b */

		/*
		 * Interrupt on all characters, errors and external int 
		 */
		writescc(hp->cardbase, cmd, R1, (INT_ALL_Rx | EXT_INT_ENAB));
#ifdef FIFO

		/*
		 * Select R7 prime 
		 */
		writescc(hp->cardbase, cmd, R15, ESCC);

		/*
		 * Set RX & TX int FIFO level
		 *
		 * Default: TX interrupt when FIFO empty
		 *          RX interrupt when FIFO half-full 
		 */
		writescc(hp->cardbase, cmd, R7, rxfifolevel | txfifolevel);
#endif
	}
#ifdef FIFO_STATUS
	writescc(hp->cardbase, cmd, R15, BRKIE | FIFOEN);       /* allow ABORT int */
#else
	writescc(hp->cardbase, cmd, R15, BRKIE);        /* allow ABORT int */
#endif

	/*
	 * Turn on the receiver and hunt for a flag Use DCD-enabled RX to
	 * avoid tons of interrupts 
	 */
	writescc(hp->cardbase, cmd, R3, RxENABLE | RxCRC_ENAB | Rx8 | auto_enab);

	restore(i_state);
	return 0;
}

/*
 * Attach a GPCM ESCC interface to the system. Parameters are:
 *
 * argv[0]:  hardware type, must be "gpcmcard"
 * argv[1]:  I/O address, e.g., "0x300"
 * argv[2]:  vector, e.g.,"2"
 * argv[3]:  8 bit dma channel (1 ... 3 available)
 * argv[4]:  mode, must be: "ax25" (AX.25 UI frame format)
 * argv[5]:  interface label(s) for A & B ports, e.g., "porta,portb"
 * argv[6]:  receiver packet buffer size in bytes
 * argv[7]:  maximum transmission unit, bytes (MUST BE < IBUFSIZE)
 * argv[8]:  channel A interface speed, e.g, "38400", 0 = ext. clock
 * argv[9]:  channel B interface speed
 * argv[10]: auto enables (RX enabled by DCD), optional (defaults to ON)
 * argv[11]: Port B RX FIFO int level, optional (defaults to 4 bytes)
 * argv[12]: Port B TX FIFO int level, optional (default to empty)
 * argv[13]: First IP address, optional (defaults to Ip_addr)
 * argv[14]: Second IP address, optional (defaults to Ip_addr) 
 */
int gpcm_attach(int argc, char *argv[])
{
	struct mbuf *bp;
	extern void refiq();
	register struct iface *if_pca, *if_pcb;
	struct gpcmchan *hp;
	int dev, cmd;
	char i_state, *cp;
	int n;

	/* 
	 * replenish interrupt buffer pool (in mbuf.c)
	 */
	refiq();
	
	/* 
	 * setup delay for writescc & readscc (in gpcmvec.asm)
	 */
	gpcm_set_acc_delay();

	/* 
	 * Quick check to make sure args are good and mycall is set
	 */
	if (strcmp(argv[4], "ax25") != 0) {
		tprintf("MODE%s unknow for interface %s\n", argv[4], argv[5]);
		return -1;
	}
	if (if_lookup(argv[5]) != NULLIF) {
		tprintf("Interface %s already exists\n", argv[5]);
		return -1;
	}
	if (Mycall[0] == '\0') {
		tprintf("Set mycall first\n");
		return -1;
	}

	/*
	 * Note: each card must have a unique address, IRQ and DMA
	 */
	if (gpcmnbr >= GPCMMAX) {
		tprintf("Maximum of %d GPCM-ESCC card supported\n", GPCMMAX);
		return -1;
	}
	dev = gpcmnbr++;

	/*
	 * Initialize hardware-level control structure 
	 */
	GPCM[dev].addr = htoi(argv[1]);
	GPCM[dev].vec = htoi(argv[2]);

	/*
	 * Set up 8253/8254 counter 0 to mode 3, 4 KHz square wave (250 us
	 * timer resolution). DIV0 is defined in gpcmescc.h 
	 */
	outportb(GPCM[dev].addr + TMRCMD, SC0 | LSB_MSB | MODE3);
	for (n = 0; n < 5; n++);/* satisfy 8253 access time restriction */

	outportb(GPCM[dev].addr + TMR0, DIV0 & 0xFF);   /* Divisor low byte */
	for (n = 0; n < 5; n++);/* satisfy 8253 access time restriction */

	outportb(GPCM[dev].addr + TMR0, DIV0 >> 8);     /* Divisor high byte */
	for (n = 0; n < 5; n++);/* satisfy 8253 access time restriction */

	/*
	 * Save original interrupt vector 
	 */
	GPCM[dev].oldvec = getirq(GPCM[dev].vec);

	/*
	 * Set new interrupt vector 
	 */
	if (setirq(GPCM[dev].vec, gpcmhandle[dev]) == -1) {
		tprintf("IRQ %u out of range\n", GPCM[dev].vec);
		gpcmnbr--;
		return -1;
	}
	if ((atoi(argv[3]) < 1) || (atoi(argv[3]) > 3)) {
		tprintf("DMA %d out of range\n", atoi(argv[3]));
		gpcmnbr--;
		return -1;
	}

	/*
	 * Set autoenable 
	 */
	if ((argc > 10) && (tolower(*argv[10]) == 'n'))
		auto_enab = 0;
	/*
	 * Set RX FIFO int level 
	 */
	if ((argc > 11) && (tolower(*argv[11]) == 'n'))
		rxfifolevel = 0;
	/*
	 * Set TX FIFO int level 
	 */
	if ((argc > 12) && (tolower(*argv[12]) == 'n'))
		txfifolevel = 0;

	/*
	 * Create interface structures and fill in details 
	 */
	if_pca = (struct iface *) callocw(1, sizeof(struct iface));
	if_pcb = (struct iface *) callocw(1, sizeof(struct iface));

	if_pca->addr = if_pcb->addr = Ip_addr;
	if (argc > 13)
		if_pca->addr = resolve(argv[13]);
	if (argc > 14)
		if_pcb->addr = resolve(argv[14]);

	if (if_pca->addr == 0 || if_pcb->addr == 0) {
		tprintf("No IP address\n");
		free((char *) if_pca);
		free((char *) if_pcb);
		return -1;
	}

	/*
	 * Set interface names 
	 */
	if ((cp = strchr(argv[5], ',')) == NULLCHAR) {

		/*
		 * Single name given: append "a" to interface associated with
		 * A channel and "b" to the one associated with B channel 
		 */
		if_pca->name = malloc((unsigned) strlen(argv[5]) + 2);
		strcpy(if_pca->name, argv[5]);
		strcat(if_pca->name, "a");
		if_pcb->name = malloc((unsigned) strlen(argv[5]) + 2);
		strcpy(if_pcb->name, argv[5]);
		strcat(if_pcb->name, "b");
	} else {

		/*
		 * Two names given: use 1st for channel A and 2nd for channel B
		 */
		*cp++ = '\0';   /* terminate 1st, cp points to 2nd */
		if_pca->name = strdup(argv[5]);
		if_pcb->name = strdup(cp);
	}
	if_pcb->mtu = if_pca->mtu = atoi(argv[7]);
	if_pcb->type = if_pca->type = CL_AX25;
	if_pcb->ioctl = if_pca->ioctl = gpcm_ctl;
	if_pca->dev = 2 * dev;  /* gpcm0a */
	if_pcb->dev = 2 * dev + 1;      /* gpcm0b */
	if_pcb->stop = if_pca->stop = gpcm_stop;
	if_pcb->output = if_pca->output = ax_output;
	if_pcb->raw = if_pca->raw = gpcm_raw;

	if (strcmp(argv[4], "ax25") == 0) {

		/*
		 * Must be true, was checked at top 
		 */
		if_pcb->send = if_pca->send = ax_send;
		if (if_pcb->hwaddr == NULLCHAR)
			if_pcb->hwaddr = mallocw(AXALEN);
		memcpy(if_pcb->hwaddr, Mycall, AXALEN);
		if (if_pca->hwaddr == NULLCHAR)
			if_pca->hwaddr = mallocw(AXALEN);
		memcpy(if_pca->hwaddr, Mycall, AXALEN);
		setencap(if_pca, "AX25");
		setencap(if_pcb, "AX25");
	}

	/*
	 * Link em in to the interface chain 
	 */
	if_pca->next = if_pcb;
	if_pcb->next = Ifaces;
	Ifaces = if_pca;

	/*
	 * set params in gpcmchan table for CHANNEL B 
	 */
	hp = &GPCMchan[2 * dev + 1];    /* gpcm1 is offset 1 */
	hp->dmachan = 0;                /* Channel B does not have dma */
	hp->cardbase = GPCM[dev].addr;
	hp->iface = if_pcb;
	hp->stata = GPCM[dev].addr + CHANA + CTL;       /* permanent status */
	hp->statb = GPCM[dev].addr + CHANB + CTL;       /* addrs for CHANA/B */
	hp->speed = atol(argv[9]);
	hp->base = GPCM[dev].addr + CHANB;
	hp->bufsiz = atoi(argv[6]);
	hp->tstate = IDLE;

	/*
	 * default channel B access Params 
	 */
	hp->txdelay = 50;       /* 50 ms */
	hp->persist = 128;      /* 50% persistence */
	hp->slotime = 100;      /* 100 ms */
	hp->squeldelay = 5;     /* 5 ms */

	/*
	 * Hardware reset the ESCC (both channel A & B at once) 
	 */
	writescc(hp->cardbase, hp->stata, R9, FHWRES);

	/*
	 * Disable interrupts with Master interrupt ctrl reg 
	 */
	writescc(hp->cardbase, hp->stata, R9, 0);

	/*
	 * check for ESCC (REQUIRED !) 
	 */
	cmd = hp->base + CTL;
	writescc(hp->cardbase, cmd, R15, 0);
	have_scc = ((readscc(hp->cardbase, cmd, R15) & ESCC) == 0);
	if (have_scc) {
		writescc(hp->cardbase, cmd, R15, ESCC);
		have_escc = (readscc(hp->cardbase, cmd, R15) & ESCC);
		writescc(hp->cardbase, cmd, R15, 0);
	} else {
		tprintf("Error: no chip found at 0x%x\n", hp->base);
		gpcmnbr--;
		return -1;
	}

	if (have_escc == 0) {
		tprintf("Error: standard SCC found at 0x%x\n", hp->base);
		gpcmnbr--;
		return -1;
	}

	/*
	 * Init ESCC channel B 
	 */
	scc_init(hp);

	/*
	 * Pre-allocate a receive buffer 
	 */
	i_state = dirps();
	hp->rcvbuf = alloc_mbuf(hp->bufsiz + sizeof(struct phdr));
	restore(i_state);
	if (hp->rcvbuf == NULLBUF) {
		/* No memory, abort receiver */
		tprintf("No memory available for recive buffer\n");
		/* Restore original interrupt vector */
		setirq(GPCM[dev].vec, GPCM[dev].oldvec);
		gpcmnbr--;
		return -1;
	}
	hp->rcvbuf->data += sizeof(struct phdr);
	hp->rcp = hp->rcvbuf->data;
	hp->rcvbuf->cnt = 0;
	hp->sndq = NULLBUF;


	/*
	 * Set params in gpcmchan table for CHANNEL A 
	 */
	hp = &GPCMchan[2 * dev];
	hp->dmachan = (unsigned char) atoi(argv[3]);

	/*
	 * Figure out where the dma page register is. 
	 */
	if (hp->dmachan < 8 && hp->dmachan >= 0) {
		hp->page_addr = Page_regs[hp->dmachan];
	} else {
		printf("DMA channel %d out of range\n", hp->dmachan);
		free_p(hp->rcvbuf);
		/* Restore original interrupt vector */
		setirq(GPCM[dev].vec, GPCM[dev].oldvec);
		gpcmnbr--;
		return -1;
	}

	hp->dma_dest = hp->dmachan * 2;
	hp->dma_wcr = hp->dma_dest + 1;

	hp->cardbase = GPCM[dev].addr;
	hp->iface = if_pca;
	hp->speed = atol(argv[8]);
	hp->base = GPCM[dev].addr + CHANA;
	hp->bufsiz = atoi(argv[6]);
	hp->tstate = IDLE;

	/*
	 * default channel access Params 
	 */
	hp->txdelay = 50;       /* 50 mS */
	hp->persist = 128;      /* 50% persistence */
	hp->slotime = 100;      /* 100 mS */
	hp->squeldelay = 5;     /* 5 mS */
#ifdef RECPROC

	/*
	 * If user allocates enough interrupt buffers to avoid ibuffails,
	 * IMHO buf_recover is not required. If you think that's wrong,
	 * #define RECPROC at top of this file. 
	 */
	newproc("buf_recover", 256, gpcm_buf_recover, 0, hp, NULL, 0);
#endif

	/*
	 * Pre-allocate a receive buffer buffer is allocated with ints off 
	 */
	i_state = dirps();
	hp->rcvbuf = alloc_mbuf(hp->bufsiz + sizeof(struct phdr));
	restore(i_state);
	if (hp->rcvbuf == NULLBUF) {

		/*
		 * No memory, abort receiver 
		 */
		tprintf("No memory available for recive buffer\n");

		/*
		 * Restore original interrupt vector 
		 */
		setirq(GPCM[dev].vec, GPCM[dev].oldvec);
		gpcmnbr--;
		return -1;
	}
	hp->rcvbuf->data += sizeof(struct phdr);
	hp->rcvbuf->cnt = 0;
	hp->sndq = NULLBUF;

	/*
	 * Get a buffer for tx which does not cross a dma boundary buffer is
	 * allocated with ints off 
	 */
	i_state = dirps();
	bp = alloc_mbuf(hp->bufsiz);
	if (bp != NULLBUF)
		hp->txdmabuf = bp->data;
	else
		hp->txdmabuf = NULLCHAR;
	restore(i_state);
	if (!hp->txdmabuf)
		tprintf("No memory available for transmit buffer\n");

	/*
	 * Init ESCC port A 
	 */
	scc_init(hp);

	/*
	 * master interrupt enable 
	 */
	writescc(hp->cardbase, CTL + hp->base, R9, MIE | NV);

	/*
	 * Enable interrupt 
	 */
	maskon(GPCM[dev].vec);

	return 0;
}

/* Shut down interface */
int gpcm_stop(iface)
struct iface *iface;
{
	int16 dev;
	struct gpcmchan *hp;

	dev = iface->dev;
	if (dev & 1)            /* Because there are 2 devices per card */
		return 0;
	dev >>= 1;              /* Convert back into gpcm number */
	hp = &GPCMchan[2 * dev];/* gpcm0a is offset 0 */

	outportb(DMA_MASK, DMA_DISABLE | hp->dmachan);  /* Disable DMA channel */

	/* Turn off interrupts */
	maskoff(GPCM[dev].vec);

	/* Restore original interrupt vector */
	setirq(GPCM[dev].vec, GPCM[dev].oldvec);

	/* Force hardware reset */
	writescc(hp->cardbase, CTL + GPCM[dev].addr + CHANA, R9, FHWRES);

	return 0;
}

/* Send raw packet on GPCM ESCC card */
int gpcm_raw(struct iface * iface, struct mbuf * bp)
{
	char i_state;
	char kickflag;
	struct gpcmchan *hp;

	dump(iface, IF_TRACE_OUT, CL_AX25, bp);
	iface->rawsndcnt++;
	iface->lastsent = secclock();

	hp = &GPCMchan[iface->dev];
	kickflag = (hp->sndq == NULLBUF) & (hp->sndbuf == NULLBUF);
	enqueue(&hp->sndq, bp);
	hp->enqueued++;
	if (kickflag) {
		/* simulate interrupt to xmit */
		switch (hp->base & 2) {
			case 2:
				a_txint(hp);    /* process interrupt */
				break;
			case 0:
#ifdef notdef
				i_state = dirps();
#endif
				if (hp->tstate == IDLE)
					b_txint(hp);
#ifdef notdef
				restore(i_state);
#endif
				break;
		}
	}
	return 0;
}

/* display GPCM ESCC card channel stats */
int dogpcmstat(int argc, char *argv[], void *p)
{
	struct gpcmchan *hp;
	int i;

	tprintf("Board Statistic:\n\n");
	tprintf("Base Addr  Rxints  Txints  Exints  TxFrms  RxFrms  Crcerr  RxOvrs  TxUndr\n");
	tprintf("---------  ------  ------  ------  ------  ------  ------  ------  ------\n");
	for (i = 0; i < gpcmnbr * 2; i++) {
		hp = &GPCMchan[i];

		tprintf("0x%03x     % 8lu% 8lu% 8lu% 8u% 8u% 8u% 8u% 8u\nRev State=%s",
		 hp->base, hp->rxints, hp->txints, hp->exints, hp->enqueued,
			hp->rxframes, hp->crcerr, hp->rovers, hp->tunders,
			hp->rstate == 0 ?
			"IDLE" : hp->rstate == 1 ?
			"ACTIVE" : hp->rstate == 2 ?
			"RXERROR" : hp->rstate == 3 ?
			"RXABORT" : "TOOBIG"
			);

		tprintf("Tstate = %s\n",
			hp->tstate == 0 ?
			"IDLE" : hp->tstate == 1 ?
			"ACTIVE" : hp->tstate == 2 ?
			"UNDERRUN" : hp->tstate == 3 ?
			"FLAGOUT" : hp->tstate == 4 ?
			"DEFER" : hp->tstate == 5 ?
			"TXDELAY" : "CRCOUT"
			);
	}
	return 0;
}

/* Subroutine to set kiss params in channel tables */
int32
gpcm_ctl(iface, cmd, set, val)
struct iface *iface;
int cmd;
int set;
int32 val;
{
	struct gpcmchan *hp;
	int32 t, ca;

	hp = &GPCMchan[iface->dev];     /* point to channel table */
	switch (cmd) {
		case PARAM_TXDELAY:
			if (set)
				hp->txdelay = val;
			return hp->txdelay;
		case PARAM_PERSIST:
			if (set)
				hp->persist = val;
			return uchar(hp->persist);
		case PARAM_SLOTTIME:
			if (set)
				hp->slotime = val;
			return hp->slotime;
		case PARAM_TXTAIL:
			if (set)
				hp->squeldelay = val;
			return hp->squeldelay;
	}
	return -1;
}


#ifdef RECPROC

/*
 * Process to recover from ibuffails. This could be done in the function
 * network() in config.c, to save a context switch.  I put it here so the
 * driver would be more self contained. 
 *
 * It is *really* needed with properly adjusted nibufs and ibufsize ??
 */
void gpcm_buf_recover(int unused, void *b, void *a)
{
	struct gpcmchan *hp0, *hp1;
	char i_state;
	int i;

	for (;;) {
		pwait(NULL);

		for (i = 0; i < gpcmnbr; i++) { /* for each card */
			hp0 = &GPCMchan[2 * i];
			hp1 = &GPCMchan[2 * i + 1];
			if (!hp0->rcvbuf) {     /* No rx buffer allocated */
				i_state = dirps();
				hp0->rcvbuf = alloc_mbuf(hp0->bufsiz + sizeof(struct phdr));
				if (hp0->rcvbuf != NULLBUF)
					hp0->rcvbuf->data += sizeof(struct phdr);
				restore(i_state);
				setup_rx_dma(hp0);
			}
			i_state = dirps();
			if (!hp1->rcvbuf && (hp1->rstate == ACTIVE)) {
				/* No rx buf allocated */
				if ((hp1->rcvbuf = alloc_mbuf(hp1->bufsiz + sizeof(struct phdr))) != NULL) {
					hp1->rcvbuf->data += sizeof(struct phdr);
					hp1->rcp = hp1->rcvbuf->data;
					hp1->rcvbuf->cnt = 0;
					writescc(hp->cardbase1, CTL + hp1->base, R3, RxENABLE | RxCRC_ENAB | Rx8);
				}
			}
			restore(i_state);
		}
	}
}

#endif
